@@ -8,16 +8,13 @@ from copy import deepcopy
 
 import ipywidgets as widgets
 import numpy as np
-import yaml
-
 import plotly.graph_objects as go
-from ipyleaflet import (DrawControl, FullScreenControl, Map, MapStyle, Polygon,
-                        WidgetControl)
+import yaml
+from ipyleaflet import DrawControl, FullScreenControl, Map, MapStyle, Polygon, WidgetControl
 
 from . import utils
-from ._leaflet import (Circle, ColorizableTileLayer, Graticule,
-                       KeyBindingControl, LayersControl, StatusBarControl,
-                       allowed_colormaps)
+from ._leaflet import (Circle, ColorizableTileLayer, Graticule, KeyBindingControl, LayersControl,
+                       StatusBarControl, allowed_colormaps)
 from .pstools import compute_ps
 
 
@@ -70,10 +67,10 @@ class App:
 
     def show_map(self):
         if self.map_config.get("use_sidecar", True):
-            from sidecar import Sidecar
             from IPython.display import display
+            from sidecar import Sidecar
 
-            sc = Sidecar(title="CMB map")
+            sc = Sidecar(title=self.map_config.get("title", "CMB map"))
             with sc:
                 display(self.m)
         else:
@@ -85,11 +82,11 @@ class App:
     def _add_layers(self):
         layers = self.map_config.get("layers", {})
         tile_default = dict(
-            min_zoom=-5,
-            max_zoom=+5,
-            min_native_zoom=-5,
-            max_native_zoom=0,
-            tile_size=675,
+            min_zoom=self.map_config.get("min_zoom", -5),
+            max_zoom=self.map_config.get("max_zoom", +5),
+            min_native_zoom=self.map_config.get("min_native_zoom", -5),
+            max_native_zoom=self.map_config.get("max_native_zoom", 0),
+            tile_size=self.map_config.get("tile_size", 675),
             show_loading=self.map_config.get("show_loading", True),
         )
         self.tiles, self.keybindings = utils.get_tiles(layers)
@@ -109,11 +106,11 @@ class App:
                 KeyBindingControl(keybindings=default_keybindings),
             ),
             crs=dict(name="CAR", custom=False),
-            center=(0, 0),
-            min_zoom=-5,
-            max_zoom=+5,
+            center=self.map_config.get("center", (0, 0)),
+            min_zoom=self.map_config.get("min_zoom", -5),
+            max_zoom=self.map_config.get("max_zoom", +5),
+            zoom=self.map_config.get("initial_zoom", 0),
             interpolation="nearest",
-            zoom=0,
             scroll_wheel_zoom=True,
             fade_animation=False,
             world_copy_jump=True,
@@ -178,7 +175,9 @@ class App:
 
         if widgets_config.get("use_cmap_control", False):
             cmap = widgets.RadioButtons(
-                options=allowed_colormaps, value="planck", layout=widgets.Layout(width="100px"),
+                options=allowed_colormaps,
+                value="planck",
+                layout=widgets.Layout(width="100px"),
             )
 
             def on_cmap_change(change):
@@ -237,7 +236,7 @@ class App:
                     id=_get_section(imap, "id"),
                     name=_get_section(imap, "file"),
                     data_type=imap.get("data_type", "IQU"),
-                    cal=None,
+                    cal=imap.get("cal"),
                 )
             )
 
@@ -256,18 +255,51 @@ class App:
 
     def _add_theory(self):
         self.theory = None
-        theory_file = self.data_config.get(
-            "theory_file", "bode_almost_wmap5_lmax_1e4_lensedCls_startAt2.dat"
-        )
-        if os.path.exists(theory_file):
+        theory = self.data_config.get("theory", {})
+        model = theory.get("model", "camb")
+        if model not in ["camb", "planck"]:
+            raise ValueError("Theory model '{}' is not supported.".format(model))
+
+        if model == "camb":
+            theory_file = theory.get("file", "bode_almost_wmap5_lmax_1e4_lensedCls_startAt2.dat")
+            if not os.path.exists(theory_file):
+                return
             clth = {}
             lth, clth["TT"], clth["EE"], clth["BB"], clth["TE"] = np.loadtxt(
-                theory_file, unpack=True,
+                theory_file,
+                unpack=True,
             )
             clth["ET"] = clth["TE"]
             for spec in ["EB", "BE", "BT", "TB"]:
                 clth[spec] = np.zeros(clth["TE"].shape)
-            self.theory = {"lth": lth, "clth": clth}
+            lth = {spec: lth for spec in clth.keys()}
+
+            # Set ell range and only show allowed spectra
+            for spec in list(clth.keys()):
+                if spec not in theory.get("allow", ["TT"]):
+                    del lth[spec]
+                    del clth[spec]
+                    continue
+                lrange = np.where(
+                    (theory.get("lmin", 2) < lth[spec]) & (lth[spec] < theory.get("lmax", 1000))
+                )
+                lth[spec] = lth[spec][lrange]
+                clth[spec] = clth[spec][lrange]
+
+        elif model == "planck":
+            lth, clth, clth_minus, clth_plus = {}, {}, {}, {}
+            for k, v in theory.get("files", {}).items():
+                lth[k], clth[k], clth_minus[k], clth_plus[k], *best_fit = np.loadtxt(v, unpack=True)
+
+        self.theory = {
+            "label": theory.get("label", "theory"),
+            "color": theory.get("color", "gray"),
+            "lth": lth,
+            "clth": clth,
+        }
+
+        if model == "planck":
+            self.theory.update({"clth_minus": clth_minus, "clth_plus": clth_plus})
 
     def _add_plot(self):
         # Main
@@ -278,10 +310,16 @@ class App:
 
         # General configuration
         layout = widgets.Layout(width="auto", height="auto")
-        self.compute_1d = widgets.Checkbox(value=True, description="1D spectra", layout=layout)
-        self.compute_2d = widgets.Checkbox(value=True, description="2D spectra", layout=layout)
+        self.compute_1d = widgets.Checkbox(
+            value=self.plot_config.get("compute_1d", True), description="1D spectra", layout=layout
+        )
+        self.compute_2d = widgets.Checkbox(
+            value=self.plot_config.get("compute_2d", True), description="2D spectra", layout=layout
+        )
         self.compute_T_only = widgets.Checkbox(
-            value=False, description="Only temperature", layout=layout
+            value=self.plot_config.get("compute_only_temperature", False),
+            description="Only temperature",
+            layout=layout,
         )
         self.lmax = widgets.IntSlider(
             value=self.plot_config.get("lmax", 1000),
@@ -353,6 +391,35 @@ class App:
         self.fig_1d = go.FigureWidget(
             layout=go.Layout(height=600, template=plotly_config.get("template", "plotly_white"))
         )
+        updatemenus = list(
+            [
+                dict(
+                    active=0,
+                    type="buttons",
+                    buttons=list(
+                        [
+                            dict(
+                                label="Linear Scale",
+                                method="relayout",
+                                args=[{"yaxis.type": "linear"}],
+                            ),
+                            dict(
+                                label="Log Scale", method="relayout", args=[{"yaxis.type": "log"}]
+                            ),
+                        ]
+                    ),
+                    direction="right",
+                    pad={"r": 10, "t": 10},
+                    showactive=True,
+                    x=1.0,
+                    xanchor="right",
+                    y=1.1,
+                    yanchor="top",
+                )
+            ]
+        )
+        layout = dict(updatemenus=updatemenus, autosize=True, xaxis_title=r"$\ell$")
+        self.fig_1d.update_layout(layout)
 
         # Header
         allowed_spectra = ["TT", "TE", "TB", "ET", "BT", "EE", "EB", "BE", "BB"]
@@ -361,7 +428,13 @@ class App:
         header = widgets.HBox([self.spectra_1d, self.split_1d])
 
         # Config
-        self.use_toeplitz = widgets.Checkbox(value=False, description="Use Toeplitz approx.")
+        self.use_toeplitz = widgets.Checkbox(
+            value=self.plot_config.get("use_toeplitz_approx", False),
+            description="Use Toeplitz approx.",
+        )
+        self.use_kspace_filter = widgets.Checkbox(
+            value=self.plot_config.get("use_kspace_filter", False), description="Use kspace filter"
+        )
         self.bin_size = widgets.IntSlider(
             value=self.plot_config.get("bin_size", 40),
             min=0,
@@ -370,6 +443,8 @@ class App:
             description="Bin size",
         )
         config = widgets.HBox([widgets.VBox([self.use_toeplitz])])
+        if self.data_config.get("filter"):
+            config.children += (widgets.VBox([self.use_kspace_filter]),)
         if not self.data_config.get("binning_file"):
             config.children += (widgets.VBox([self.bin_size]),)
         accordion = widgets.Accordion(children=[config], selected_index=None)
@@ -384,6 +459,54 @@ class App:
         self.fig_2d = go.FigureWidget(
             layout=go.Layout(height=600, template=plotly_config.get("template", "plotly_white"))
         )
+        updatemenus = list(
+            [
+                dict(
+                    active=0,
+                    buttons=list(
+                        [
+                            dict(
+                                label="2D",
+                                method="restyle",
+                                args=["type", "heatmap"],
+                            ),
+                            dict(
+                                label="3D",
+                                method="restyle",
+                                args=["type", "surface"],
+                            ),
+                        ]
+                    ),
+                    direction="down",
+                    pad={"r": 10, "t": 10},
+                    showactive=True,
+                    x=1.0,
+                    xanchor="right",
+                    y=1.15,
+                    yanchor="top",
+                )
+            ]
+        )
+        layout = dict(
+            updatemenus=updatemenus,
+            # sliders=sliders,
+            autosize=True,
+            # width=self.fig_2d.layout.height,
+            xaxis=dict(
+                title=r"$\ell_X$",
+                showgrid=False,
+                zeroline=False,
+                constrain="domain",
+            ),
+            yaxis=dict(
+                title=r"$\ell_Y$",
+                scaleanchor="x",
+                scaleratio=1,
+                showgrid=False,
+                zeroline=False,
+            ),
+        )
+        self.fig_2d.update_layout(layout)
 
         # Header
         allowed_spectra = ["II", "IQ", "IU", "QI", "QQ", "QU", "UI", "UQ", "UU"]
@@ -411,10 +534,13 @@ class App:
         ):
             if not compute:
                 continue
-            for ipatch, (name, patch) in enumerate(self.patches.items()):
-                print("Compute patch #{} for '{}' method".format(ipatch, ps_method))
+            for name, patch in self.patches.items():
+                print("Compute {} for '{}' method".format(name, ps_method))
 
-                kwargs = dict(ps_method=ps_method, lmax=self.lmax.value,)
+                kwargs = dict(
+                    ps_method=ps_method,
+                    lmax=self.lmax.value,
+                )
                 if ps_method == "master":
                     kwargs.update(
                         dict(
@@ -430,6 +556,15 @@ class App:
                             l_toep=2500 if self.use_toeplitz.value else None,
                         )
                     )
+                    if self.use_kspace_filter.value:
+                        filter_config = self.data_config.get("filter")
+                        kwargs.update(
+                            dict(
+                                vk_mask=filter_config.get("vk_mask"),
+                                hk_mask=filter_config.get("hk_mask"),
+                                transfer_function=filter_config.get("transfer_function"),
+                            )
+                        )
 
                 method = patch.get(ps_method, dict())
                 if method.get("results") and method.get("config") == kwargs:
@@ -467,14 +602,15 @@ class App:
                     self.split_2d.options = spec_name_list
 
         if spectra is not None:
-            self.spectra_1d.observe(self._update_plot, names="value")
-            self.split_1d.observe(self._update_plot, names="value")
-            self.spectra_2d.observe(self._update_plot, names="value")
-            self.split_2d.observe(self._update_plot, names="value")
+            self.spectra_1d.observe(self._update_1d_plot, names="value")
+            self.split_1d.observe(self._update_1d_plot, names="value")
+            self.spectra_2d.observe(self._update_2d_plot, names="value")
+            self.split_2d.observe(self._update_2d_plot, names="value")
             self.patch_2d.options = self.patches.keys()
-            self.patch_2d.observe(self._update_plot, names="value")
+            self.patch_2d.observe(self._update_2d_plot, names="value")
 
-            self._update_plot(None)
+            self._update_1d_plot(None, create=True)
+            self._update_2d_plot(None, create=True)
 
         self.compute_button.description = "Compute spectra"
         self.compute_button.icon = "check"
@@ -482,20 +618,50 @@ class App:
         self.clean_button.disabled = False
         self.export_button.disabled = False
 
-    def _update_plot(self, _):
-        if self.compute_1d.value:
-            self._update_1d_plot()
-        if self.compute_2d.value:
-            self._update_2d_plot()
-
-    def _update_1d_plot(self):
+    def _update_1d_plot(self, _, create=False):
         split_name = self.split_1d.value
         spec = self.spectra_1d.value
 
-        # Clean data
-        self.fig_1d.data = []
+        if create:
+            # Clean data
+            self.fig_1d.data = []
 
         # Update theory & data
+        if self.theory is not None:
+            x, y = [], []
+            yerr_minus, yerr_plus = [], []
+            if spec in self.theory.get("clth"):
+                x = self.theory.get("lth").get(spec)
+                y = self.theory.get("clth").get(spec)
+                yerr_minus = (
+                    self.theory.get("clth_minus").get(spec) if "clth_minus" in self.theory else []
+                )
+                yerr_plus = (
+                    self.theory.get("clth_plus").get(spec) if "clth_plus" in self.theory else []
+                )
+            if create:
+                self.fig_1d.add_scatter(
+                    name=self.theory.get("label"),
+                    x=x,
+                    y=y,
+                    mode="markers" if "clth_minus" in self.theory else "lines",
+                    line=dict(color=self.theory.get("color")),
+                    error_y=dict(
+                        type="data",
+                        symmetric=False,
+                        array=yerr_plus,
+                        arrayminus=yerr_minus,
+                        color=self.theory.get("color"),
+                    ),
+                )
+            else:
+                with self.fig_1d.batch_update():
+                    self.fig_1d.data[0].x = x
+                    self.fig_1d.data[0].y = y
+                    self.fig_1d.data[0].error_y.array = yerr_plus
+                    self.fig_1d.data[0].error_y.arrayminus = yerr_minus
+
+        idata = 0 if self.theory is None else 1
         for name, patch in self.patches.items():
             method = patch.get("master", dict())
             results = method.get("results")
@@ -506,81 +672,31 @@ class App:
             x = results.get("lb")
             y = results.get("ps").get(split_name).get(spec)
             yerr = np.sqrt(np.diag(results.get("cov").get(split_name).get(spec)))
-            self.fig_1d.add_scatter(
-                name=name,
-                x=x,
-                y=y,
-                error_y=dict(type="data", array=yerr, visible=True, color=color),
-                mode="markers",
-                marker=dict(color=color),
-            )
-        if self.theory is not None:
-            self.fig_1d.add_scatter(
-                name="theory",
-                x=self.theory.get("lth")[: self.lmax.value],
-                y=self.theory.get("clth")[self.spectra_1d.value][: self.lmax.value],
-                mode="lines",
-                line=dict(color="gray"),
-            )
-
-        yaxis_title = r"$D_\ell^\mathrm{%s}$" % spec
-        updatemenus = list(
-            [
-                dict(
-                    active=0,
-                    type="buttons",
-                    buttons=list(
-                        [
-                            dict(
-                                label="Linear Scale",
-                                method="update",
-                                args=[
-                                    {"visible": [True] * len(self.fig_1d.data)},
-                                    {"yaxis": {"title": yaxis_title, "type": "linear"}},
-                                ],
-                            ),
-                            dict(
-                                label="Log Scale",
-                                method="update",
-                                args=[
-                                    {"visible": [True] * len(self.fig_1d.data)},
-                                    {"yaxis": {"title": yaxis_title, "type": "log"}},
-                                ],
-                            ),
-                        ]
-                    ),
-                    direction="right",
-                    pad={"r": 10, "t": 10},
-                    showactive=True,
-                    x=1.0,
-                    xanchor="right",
-                    y=1.1,
-                    yanchor="top",
+            if create:
+                self.fig_1d.add_scatter(
+                    name=name,
+                    x=x,
+                    y=y,
+                    error_y=dict(type="data", array=yerr, visible=True, color=color),
+                    mode="markers",
+                    marker=dict(color=color),
                 )
-            ]
-        )
-        layout = dict(
-            updatemenus=updatemenus,
-            autosize=True,
-            title=split_name,
-            xaxis_title=r"$\ell$",
-            yaxis_title=yaxis_title,
-        )
-        try:
-            self.fig_1d.update_layout(layout)
-        except ValueError:
-            # Do not know what's going on : an exception is raised by the graphical
-            # representation works well so skip it for the time being
-            print("Exception raised from 'update_layout'")
-            pass
+            else:
+                with self.fig_1d.batch_update():
+                    self.fig_1d.data[idata].x = x
+                    self.fig_1d.data[idata].y = y
+                    self.fig_1d.data[idata].error_y.array = yerr
+                idata += 1
 
-    def _update_2d_plot(self):
+        with self.fig_1d.batch_update():
+            self.fig_1d.layout.title = split_name
+            self.fig_1d.layout.yaxis.title = r"$D_\ell^\mathrm{%s}\;[\mu\mathrm{K}^2]$" % spec
+
+    def _update_2d_plot(self, _, create=False):
         split_name = self.split_2d.value
         spec = self.spectra_2d.value
         patch_name = self.patch_2d.value
-        method = self.patches[patch_name].get("2dflat")
-        if not method:
-            return
+        method = self.patches[patch_name].get("2dflat", dict())
         results = method.get("results")
         if not results:
             return
@@ -590,10 +706,6 @@ class App:
         x = np.linspace(np.min(p2d.lx), np.max(p2d.lx), shape[1])
         y = np.linspace(np.min(p2d.ly), np.max(p2d.ly), shape[0])
 
-        # Clean data
-        self.fig_2d.data = []
-        self.fig_2d.add_heatmap(x=x, y=y, z=powermap, colorscale=default_colorscale)
-
         # Create slider and buttons
         zmin, zmax = np.min(powermap), np.max(powermap)
         steps = []
@@ -603,12 +715,14 @@ class App:
                 method="restyle",
                 args=[
                     dict(
-                        zmin=scale * zmin, zmax=scale * zmax, cmin=scale * zmin, cmax=scale * zmax,
+                        zmin=scale * zmin,
+                        zmax=scale * zmax,
+                        cmin=scale * zmin,
+                        cmax=scale * zmax,
                     )
                 ],
             )
             steps.append(step)
-
         sliders = [
             dict(
                 len=0.5,
@@ -619,51 +733,25 @@ class App:
                 steps=steps,
             )
         ]
-        updatemenus = list(
-            [
-                dict(
-                    active=0,
-                    buttons=list(
-                        [
-                            dict(label="2D", method="restyle", args=["type", "heatmap"],),
-                            dict(label="3D", method="restyle", args=["type", "surface"],),
-                        ]
-                    ),
-                    direction="down",
-                    pad={"r": 10, "t": 10},
-                    showactive=True,
-                    x=1.0,
-                    xanchor="right",
-                    y=1.15,
-                    yanchor="top",
-                )
-            ]
-        )
-        layout = dict(
-            updatemenus=updatemenus,
-            sliders=sliders,
-            autosize=True,
-            # width=self.fig_2d.layout.height,
-            title=dict(
-                text="{} - {}".format(split_name, patch_name),
-                font=dict(
-                    color=self.patches[patch_name].get("properties").get("style").get("color")
-                ),
-            ),
-            xaxis=dict(
-                title=r"$\ell_X$",
-                showgrid=False,
-                zeroline=False,
-                range=[-self.lmax.value, +self.lmax.value],
-                constrain="domain",
-            ),
-            yaxis=dict(
-                title=r"$\ell_Y$",
-                scaleanchor="x",
-                scaleratio=1,
-                showgrid=False,
-                zeroline=False,
-                range=[-self.lmax.value, +self.lmax.value],
-            ),
-        )
-        self.fig_2d.update_layout(layout)
+
+        if create:
+            # Clean data
+            self.fig_2d.data = []
+            self.fig_2d.add_heatmap(x=x, y=y, z=powermap, colorscale=default_colorscale)
+        else:
+            with self.fig_2d.batch_update():
+                self.fig_2d.data[0].x = x
+                self.fig_2d.data[0].y = y
+                self.fig_2d.data[0].z = powermap
+                self.fig_2d.data[0].zmin = zmin
+                self.fig_2d.data[0].zmax = zmax
+
+        with self.fig_2d.batch_update():
+            self.fig_2d.layout.sliders = sliders
+            self.fig_2d.layout.title.text = "{} - {}".format(split_name, patch_name)
+            self.fig_2d.layout.title.font.color = (
+                self.patches[patch_name].get("properties").get("style").get("color")
+            )
+            self.fig_2d.layout.xaxis.range = [-self.lmax.value, +self.lmax.value]
+            self.fig_2d.layout.yaxis.range = [-self.lmax.value, +self.lmax.value]
+        self.fig_2d.update_traces()
